#include <WiFi.h>          
#include <DNSServer.h>
#include <WebServer.h>
#include <WiFiManager.h>          //　Install WiFiManager by tzapu using Library Manager of Arduino IDE.

// NTP settings
#define NTP_SERVER "pool.ntp.org"
// find the code of timezone from https://github.com/nayarsystems/posix_tz_db/blob/master/zones.csv
#define TIMEZONE "JST-9"

// for analog priority mode, set true
#define PRIORITY_ANALOG false

#if PRIORITY_ANALOG // analog priority mode : analog clock has 1min resolution
  #define MIN_STEP 1 // minute hand has 1 min resolution
  #define HOUR_STEP_IN_MIN 12 // hour hand moves every 12 minutes
#else // digital priority mode : digits are always aligned
  #define MIN_STEP 10 // minute hand has 10 min resolution
  #define HOUR_STEP_IN_MIN 60  // hour hand moves once a hour
#endif

//////////////////////////////
// motor control section
//////////////////////////////

// Motor parameters
#define STEPS_PER_ROTATION 4096 // steps of a single rotation of motor
#define MOTOR_DELAY 1800
#define KILL_BACKLASH 13
#define INITIAL_POS 33

// ports used to control the stepper motor
// if your motor rotates to the opposite direction, 
// change to int port[4] = {1, 2, 21, 22};
int port[4] = {22, 21, 2, 1}; // Seeed Studio XIAO ESP32C6

// sequence of stepper motor control
int seq[8][4] = {
  {  LOW,  LOW, HIGH,  LOW},
  {  LOW,  LOW,  LOW, HIGH},
  { HIGH,  LOW,  LOW,  LOW},
  {  LOW, HIGH,  LOW,  LOW}
};

// stepper motor rotation
void rotateSimple(int step) {
  int count = 0;
  static int phase = 0;
  int i, j;
  int delta = (step > 0) ? 1 : 3;

  step = (step > 0) ? step : -step;
  for(j = 0; j < step; j++) {
    phase = (phase + delta) % 4;
    for(i = 0; i < 4; i++) {
      digitalWrite(port[i], seq[phase][i]);
    }
    count++;
    delayMicroseconds(MOTOR_DELAY);
    if(step - j < 100) delayMicroseconds(MOTOR_DELAY); // deacceleration
    if(step - j < 200) delayMicroseconds(MOTOR_DELAY); // deacceleration
    if(step - j < 400) delayMicroseconds(MOTOR_DELAY); // deacceleration
    if(count < 100) delayMicroseconds(MOTOR_DELAY); // acceleration
    if(count < 50) delayMicroseconds(MOTOR_DELAY); // acceleration
    if(count < 20) delayMicroseconds(MOTOR_DELAY); // acceleration
    if(count < 10) delayMicroseconds(MOTOR_DELAY); // acceleration
  }
  // power cut
  for(i = 0; i < 4; i++) {
    digitalWrite(port[i], LOW);
  }
}

void rotate(int step) {
  static int prevStep = 0;

  if(step < 0 && prevStep < 0) { // same direction, negative
    rotateSimple(KILL_BACKLASH); // move forward to make approach run space
    rotateSimple(step - KILL_BACKLASH);
  }
  else if(step > 0 && prevStep > 0) { // same direction, positive
    rotateSimple(-KILL_BACKLASH); // move backward to make approach run space
    rotateSimple(step + KILL_BACKLASH);
  }
  else { // switchback
    rotateSimple(step);
  }
  prevStep = step;
}

//////////////////////////////
// rotary clock control section
//////////////////////////////

#define POSITION 60 // GCM of 12, 6, and 10
#define DIGIT 3

typedef struct {
  int v[DIGIT];
} Digit;

void printDigit(Digit d);
Digit rotUp(Digit current, int digit, int num);
Digit rotDown(Digit current, int digit, int num);
Digit rotDigit(Digit current, int digit, int num);
Digit setDigit(Digit current, int digit, int num);
int setNumber(Digit n);

// avoid error accumuration of fractional part of 4096 / POSITION
void rotStep(int s) {
  static long currentPos;
  static long currentStep;
  
  currentPos += s;
  long diff = currentPos * STEPS_PER_ROTATION / POSITION - currentStep;
  if(diff < 0) diff -= KILL_BACKLASH;
  else diff += KILL_BACKLASH;
  rotate(diff);
  currentStep += diff;
}

void printDigit(Digit d) {
  String s;
  int i;

  for(i = 0; i < DIGIT; i++) {
    Serial.print(String(d.v[i]));
    Serial.print(" ");
  }
    Serial.println("");

}

Digit current = {0};

//increase specified digit
Digit rotUp(Digit current, int digit, int num) {
  int freeplay = 0;
  int i;
  
  for(i = digit; i < DIGIT - 1; i++) {
    int id = current.v[i];
    int nd = current.v[i+1];
    if(id <= nd) id += POSITION;
    freeplay += id - nd - 1;
  }
  freeplay += num;
  rotStep(-1 * freeplay);
  current.v[digit] = (current.v[digit] + num) % POSITION;
  for(i = digit + 1; i < DIGIT; i++) {
    current.v[i] = (current.v[i - 1] + (POSITION - 1)) % POSITION;
  }

  Serial.print("up end : ");
  printDigit(current);
  return current;
}

// decrease specified digit
Digit rotDown(Digit current, int digit, int num) {
  int freeplay = 0;
  int i;
  
  for(i = digit; i < DIGIT - 1; i++) {
    int id = current.v[i];
    int nd = current.v[i+1];
    if(id > nd) nd += POSITION;
    freeplay += nd - id;
  }
  freeplay += num;
  rotStep( 1 * freeplay);
  current.v[digit] = (current.v[digit] - num + POSITION) % POSITION;
  for(i = digit + 1; i < DIGIT; i++) {
    current.v[i] = current.v[i - 1];
  }

  Serial.print("down end : ");
  printDigit(current);
  return current;
}

// decrease or increase specified digit 
Digit rotDigit(Digit current, int digit, int num) {
  if(num > 0) {
    return rotUp(current, digit, num);
  }
  else if(num < 0) {
    return rotDown(current, digit, -num);
  }
  else return current;
}

// set single digit to the specified number
Digit setDigit(Digit current, int digit, int num) {
  int pd, cd = current.v[digit];
  if(digit == 0) { // most significant digit
    pd = 0;
  }
  else {
    pd = current.v[digit - 1];
  }
  if(cd == num) return current;
  
  // check if increasing rotation is possible
  int n2 = num;
  if(n2 < cd) n2 += POSITION;
  if(pd < cd) pd += POSITION;
  if(pd <= cd || pd > n2) {
    return rotDigit(current, digit, n2 - cd);
  }
  // if not, do decrease rotation
  if(num > cd) cd += POSITION;
  return rotDigit(current, digit, num - cd);  
}

int setNumber(Digit n) {
  int i;
  
  for(i = 0; i < DIGIT; i++) {
    if(current.v[i] != n.v[i]) {
      break;
    }
  }
  // rotate most significant wheel only to follow current time ASAP
  if(i < DIGIT) {
      current = setDigit(current, i, n.v[i]);
  }
  if(i >= DIGIT - 1) {
    return true; // finished
  }
  else {
    return false;
  }
}

//////////////////////////////
// WiFi and NTP section
//////////////////////////////

void wifiSetup(void) {
  WiFiManager wifiManager;

  //reset saved settings
  //wifiManager.resetSettings();

  wifiManager.setTimeout(180);

  wifiManager.autoConnect("Hybrid Clock");

  //if you get here you have connected to the WiFi
  Serial.print("connected. local ip : ");
  Serial.println(WiFi.localIP());
  Serial.print("SSID : ");
  Serial.println(WiFi.SSID());
}

void getNTP(void) {
  for(int i = 0; WiFi.status() != WL_CONNECTED; i++) {
    if(i > 30) {
      ESP.restart();
    }
    Serial.println("Waiting for WiFi connection..");
    delay(1000);
  }

  configTime(0, 0, NTP_SERVER);
  setenv("TZ", TIMEZONE, 1);
  tzset();

  printLocalTime();
}

void printLocalTime() {
  struct tm timeinfo;
  if (!getLocalTime(&timeinfo)) {
    Serial.println("Failed to obtain time");
    return;
  }
  Serial.println(&timeinfo, "%A, %Y-%m-%d %H:%M:%S");
}

//////////////////////////////
// setup and main loop
//////////////////////////////

void setup() {
  Serial.begin(115200);
  Serial.println("start");

  pinMode(port[0], OUTPUT);
  pinMode(port[1], OUTPUT);
  pinMode(port[2], OUTPUT);
  pinMode(port[3], OUTPUT);

  wifiSetup();
  
  rotate(STEPS_PER_ROTATION * DIGIT); // reset all digits using physical end stop
  rotate(-INITIAL_POS * DIGIT); // release pushing force to align all digits better

  getNTP(); // get current time

}

void loop() {
  static int prevhour = -1;
  struct tm tmtime;
  boolean finished;

  getLocalTime(&tmtime);

  Digit n;

  n.v[0] = (tmtime.tm_hour % 12) *  5 + (tmtime.tm_min / HOUR_STEP_IN_MIN) * HOUR_STEP_IN_MIN / 12; // hour pos
  n.v[1] = (tmtime.tm_min / MIN_STEP) * MIN_STEP; // 10min pos
  n.v[2] = (tmtime.tm_min % 10) *  6; // 1min pos

  finished = setNumber(n);

  if(finished && tmtime.tm_hour != prevhour) {
    if(prevhour == 11) // 12:00
      abort();
    prevhour = tmtime.tm_hour;
  }
}
